module Recurify
  module RuleAnalysis # :nodoc:
    # Returns +true+ if +self+ is normalized. By definition, a +Rule+ is
    # normalized if and only if it is not denormalized.
    #
    # @see #denormalized?
    #
    # @return [Boolean]
    def normalized?
      attributes == normalized_attributes
    end

    # Returns +true+ if +self+ is denormalized. By definition, a +Rule+ is
    # denormalized if and only if it is not normalized.
    #
    # @see #normalized?
    #
    # @return [Boolean]
    def denormalized?
      !normalized?
    end

    # Returns +true+ if +self+ is finite. By definition, a +Rule+ is finite if
    # and only if it is not infinite.
    #
    # @see #infinite?
    #
    # @retrun [Boolean]
    def finite?
      # @todo Checking of the #count is not needed anymore, once #normalize
      #   supports minifying of #ends_on.
      !count.nil? || !normalized_ends_on.nil?
    end

    # Returns +true+ if +self+ is infinite. By definition, a +Rule+ is infinite
    # if and only if it is not finite.
    #
    # @see #finite?
    #
    # @return [Boolean]
    def infinite?
      !finite?
    end

    # Returns +true+ if and only if +self+ evaluates to a single occurrence.
    #
    # @return [Boolean]
    def one?
      starts_on == normalized_ends_on
    end

    # Returns +true+ if +self+ starts before or on +upper+. If +upper == nil+,
    # it is interpreted as "positive infinity". In that case, +#starts_before?+
    # always returns +true+, because all +Rule+ objects must have a finite
    # +#starts_on+ attribute.
    #
    # @param upper [#to_date, nil]
    #
    # @return [Boolean]
    def starts_before?(upper)
      upper.nil? || starts_on <= upper.to_date
    end

    # Returns +true+ if +self+ starts after or on +lower+. If +lower == nil+, it
    # is interpreted as "negative infinity". In that case, +#starts_after?+
    # always returns +true+, because all +Rule+ objects must have a finite
    # +#starts_on+ attribute.
    #
    # @param lower [#to_date, nil]
    #
    # @return [Boolean]
    def starts_after?(lower)
      lower.nil? || starts_on >= lower.to_date
    end

    # Returns +true+ if +self+ starts after +lower+ *and* starts before
    # +upper+ (including boundaries).
    #
    # @see #starts_before?
    # @see #starts_after?
    #
    # @param lower [#to_date, nil]
    # @param upper [#to_date, nil]
    #
    # @return [Boolean]
    def starts_between?(lower: nil, upper: nil)
      starts_after?(lower) && starts_before?(upper)
    end

    # Returns +true+ if +self+ ends before or on +upper+. If +upper == nil+, it
    # is interpreted as "positive infinity". In that case, +#ends_before?+
    # always returns +true+.
    #
    # @param upper [#to_date, nil]
    #
    # @return [Boolean]
    def ends_before?(upper)
      false ||
        # Case (1): [anything] <= 'infinite' ==> true
        upper.nil? ||
        # Case (2): 'finite' <= 'finite' ==> depends on actual comparison
        !normalized_ends_on.nil? && normalized_ends_on <= upper.to_date
    end

    # Returns +true+ if +self+ ends after or on +lower+. If +lower == nil+, it
    # is interpreted as "negative infinity". In that case, +#ends_after?+ always
    # returns +true+.
    #
    # @param lower [#to_date, nil]
    #
    # @return [Boolean]
    def ends_after?(lower)
      false ||
        # Case (1): 'infinite' >= [anything] ==> true
        normalized_ends_on.nil? ||
        # Case (2): 'finite' >= 'infinite' ==> true
        lower.nil? ||
        # Case (3): 'finite' >= 'finite' ==> depends on actual comparison
        normalized_ends_on >= lower.to_date
    end

    # Returns +true+ if +self+ ends after +lower+ *and* ends before +upper+
    # (including boundaries).
    #
    # @see #ends_before?
    # @see #ends_after?
    #
    # @param lower [#to_date, nil]
    # @param upper [#to_date, nil]
    #
    # @return [Boolean]
    def ends_between?(lower: nil, upper: nil)
      ends_after?(lower) && ends_before?(upper)
    end

    # Returns +true+ if +self+ starts *and/or* ends between +lower+ and +upper+
    # (including boundaries).
    #
    # @see #starts_between?
    # @see #ends_between?
    #
    # @param lower [#to_date, nil]
    # @param upper [#to_date, nil]
    #
    # @return [Boolean]
    def overlaps?(lower: nil, upper: nil)
      false ||
        # Case (1): +lower+ <= starts_on <= +upper+ ==> true
        starts_between?(lower: lower, upper: upper) ||
        # Case (2): +lower+ <= normalized_ends_on <= +upper+ ==> true
        ends_between?(lower: lower, upper: upper)
    end
  end
end
