require 'forwardable'

module Recurify
  module RuleTranslation # :nodoc:
    extend Forwardable

    # Applicable frequency minimizations grouped by +#frequency+ ...
    FREQUENCY_MINIMIZATION_MATRIX = {
      'daily' =>     { target_frequency: 'daily',   interval_multiplier:  1 },
      'monthly' =>   { target_frequency: 'monthly', interval_multiplier:  1 },
      'weekly' =>    { target_frequency: 'daily',   interval_multiplier:  7 },
      'quarterly' => { target_frequency: 'monthly', interval_multiplier:  3 },
      'yearly' =>    { target_frequency: 'monthly', interval_multiplier: 12 }
    }.freeze

    # Applicable frequency maximizations grouped by +#frequency+ and ordered by
    # decreasing preferability ...
    FREQUENCY_MAXIMIZATION_MATRIX = {
      'daily' => [
        { target_frequency: 'weekly',    interval_divisor:  7 },
        { target_frequency: 'daily',     interval_divisor:  1 }
      ],
      'weekly' => [
        { target_frequency: 'weekly',    interval_divisor:  1 }
      ],
      'monthly' => [
        { target_frequency: 'yearly',    interval_divisor: 12 },
        { target_frequency: 'quarterly', interval_divisor:  3 },
        { target_frequency: 'monthly',   interval_divisor:  1 }
      ],
      'quarterly' => [
        { target_frequency: 'yearly',    interval_divisor:  4 },
        { target_frequency: 'quarterly', interval_divisor:  1 }
      ],
      'yearly' => [
        { target_frequency: 'yearly',    interval_divisor:  1 }
      ]
    }.freeze

    # Creates a new +Rule+, similar to +self+, but with minimized +#frequency+.
    # Note that +#minimize_frequency+ is idempotent.
    #
    # Frequency minimization means that +Rule+s with +#frequency+ ...
    # * ... "weekly" are translated to +Rule+s with +#frequency+ "daily".
    # * ... "quarterly" are translated to +Rule+s with +#frequency+ "monthly".
    # * ... "yearly" are translated to +Rule+s with +#frequency+ "monthly".
    #
    # Additional translations may be added in the future.
    #
    # @see FREQUENCY_MINIMIZATION_MATRIX
    #
    # @return [Rule]
    def minimize_frequency
      translate_with = FREQUENCY_MINIMIZATION_MATRIX[frequency]

      substitute(
        frequency: translate_with[:target_frequency],
        interval:  interval * translate_with[:interval_multiplier]
      )
    end

    # Creates a new +Rule+, similar to +self+, but normalized. Note that
    # +#normalize+ is idempotent.
    #
    # @todo This method is not yet complete, because +#normalize+ should also
    #   minimize +#ends_on+ and +#count+ (if possible).
    #
    # @return [Rule]
    def normalize
      @_normalized_rule ||= minimize_frequency
    end

    # @!method normalized_attributes
    #   @see Rule#attributes
    #   @see #normalize
    #   @return [Hash<Symbol,Object>]
    def_delegator :normalize, :attributes, :normalized_attributes

    # @!method normalized_frequency
    #   @see Rule#frequency
    #   @see #normalize
    #   @return [String]
    def_delegator :normalize, :frequency, :normalized_frequency

    # @!method normalized_interval
    #   @see Rule#interval
    #   @see #normalize
    #   @return [Fixnum]
    def_delegator :normalize, :interval, :normalized_interval

    # @!method normalized_count
    #   @see Rule#count
    #   @see #normalize
    #   @return [Fixnum, nil]
    def_delegator :normalize, :count, :normalized_count

    # @!method normalized_ends_on
    #   @see Rule#ends_on
    #   @see #normalize
    #   @return [Date, nil]
    def_delegator :normalize, :ends_on, :normalized_ends_on

    # Creates a new +Rule+, similar to +self+, but with maximized +#frequency+.
    # Note that +#maximize_frequency+ is idempotent.
    #
    # Frequency maximization means that +Rule+s with +#frequency+ ...
    # * ... "daily" are translated to +Rule+s with +#frequency+ "weekly".
    # * ... "monthly" are translated to +Rule+s with +#frequency+ "yearly".
    # * ... "monthly" are translated to +Rule+s with +#frequency+ "quarterly".
    #
    # Additional translations may be added in the future.
    #
    # Why maximize the +#frequency+ at all? The idea is that frequency-maximized
    # +Rule+s are easier to parse for humans. For instance, "every 7th week" is
    # easier to understand than "every 49th day".
    #
    # @see FREQUENCY_MAXIMIZATION_MATRIX
    #
    # @return [Rule]
    def maximize_frequency
      translate_with = FREQUENCY_MAXIMIZATION_MATRIX[frequency].find do |c|
        interval % c[:interval_divisor] == 0
      end

      substitute(
        frequency: translate_with[:target_frequency],
        interval:  interval / translate_with[:interval_divisor]
      )
    end

    # Creates a new +Rule+, similar to +self+, but denormalized. Note that
    # +#denormalized+ is idempotent.
    #
    # @return [Rule]
    def denormalize
      @_denormalized_rule ||= maximize_frequency
    end

    # @!method denormalized_attributes
    #   @see Rule#attributes
    #   @see #denormalize
    #   @return [Hash<Symbol,Object>]
    def_delegator :denormalize, :attributes, :denormalized_attributes

    # @!method denormalized_frequency
    #   @see Rule#frequency
    #   @see #denormalize
    #   @return [String]
    def_delegator :denormalize, :frequency, :denormalized_frequency

    # @!method denormalized_interval
    #   @see Rule#interval
    #   @see #denormalize
    #   @return [Fixnum]
    def_delegator :denormalize, :interval, :denormalized_interval

    # @!method denormalized_count
    #   @see Rule#count
    #   @see #denormalize
    #   @return [Fixnum, nil]
    def_delegator :denormalize, :count, :denormalized_count

    # @!method denormalized_ends_on
    #   @see Rule#ends_on
    #   @see #denormalize
    #   @return [Date, nil]
    def_delegator :denormalize, :ends_on, :denormalized_ends_on
  end
end
